from datto.CleanText import CleanText
from datto.Experiments import Experiments
from datto.DataConnections import SQLConnections
from datto.DataConnections import S3Connections
from datto.SetupMethods import SetupMethods
from datto.TrainModel import TrainModel
from datto.ModelResults import ModelResults
from datto.FrontEnd import FrontEnd
from datto.Eda import Eda



import hashlib

import numpy as np


class Experiments:
    def assign_condition_by_id(
        self, user_id, conditions, proportions_by_conditions, random_state
    ):
        """
        Assign a given id to the same experimental condition every time for a consistent user experience.
        I.e. customer #15 will always be in the treatment condition.

        Parameters
        --------
        id: int
        conditions: numpy array
            E.g. ['treatment', 'control']
        proportions_by_conditions: numpy array
            Should add up to 1, e.g. [0.8, 0.2]
        random_state: int
            Divisor used for consistent assignment

        Returns
        --------
        chosen_condition: str
            Chooses one of the conditions you provided

        """
        assert len(conditions) == len(
            proportions_by_conditions
        ), "Need a proportion of assignment for each condition (and vice versa)."

        assert (
            np.array(proportions_by_conditions).sum() == 1.0
        ), "Need proportions to add up to 1."

        md5_result = hashlib.md5(str(user_id).encode())
        hex_string = md5_result.hexdigest()

        # Each hexadecimal character carries 4 bits of information.
        # The integers in Python are 32 bits or 64 bits depending on system architecture.
        # To be safe, we'll assume a 32 bit architecure, even though it is almost certainly 64 bits.
        # That means we can process only 8 characters of hex into int without fear of losing fidelity.
        hex_string_truncated = hex_string[-8:]

        # Hexadecimal is a base 16 representation, so convert the hex characters to integers.
        numeric_result = int(hex_string_truncated, 16)

        # From the numeric, select a condition.
        # We force this integer between 0 and random_state-1 via the modulo.
        remainder = numeric_result % random_state
        thresholds_for_condition_assignment = np.floor(
            np.array(proportions_by_conditions).cumsum() * random_state
        )

        # Get the first index where the remainder is less than the condition boundry
        condition_index = np.where(remainder < thresholds_for_condition_assignment)[0][
            0
        ]
        chosen_condition = np.array(conditions)[condition_index]

        return chosen_condition



class FrontEnd:
    def dropdown_from_dataframe(self, name, df, chosen_col):
        """
        Create text to use for rendering an HTML dropdown from a DataFrame.

        Render by using {{ df|safe }} in your HTML file.

        Parameters
        --------
        name: str
            Name you'd like for the dropdown
        df: DataFrame
        chosen_col: str
            Which column's values will populate the dropdown

        Returns
        --------
        html_choices: str
            String you can use to render HTML

        """
        html_choices = """<select id="{}" width: 400px;><option value="---">---</option>""".format(
            name
        )
        for option in df[chosen_col].unique():
            html_choices += """<option value="{0}">{0}</option>""".format(option)
        html_choices += """</select>"""

        return html_choices



import re
import string
from math import ceil
from operator import itemgetter
from random import randrange

import lime
import lime.lime_tabular
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import shap
from gensim.corpora import Dictionary
from gensim.models import CoherenceModel, nmf
from sklearn.decomposition import NMF
from sklearn.feature_extraction.text import (
    ENGLISH_STOP_WORDS,
    CountVectorizer,
    TfidfVectorizer,
)
from sklearn.linear_model import ElasticNet, LogisticRegression, SGDClassifier
from sklearn.tree import DecisionTreeClassifier
from sklearn.metrics import (
    accuracy_score,
    mean_squared_error,
    median_absolute_error,
    precision_score,
    r2_score,
    recall_score,
    roc_auc_score,
    f1_score
)
from sklearn.model_selection import train_test_split
from sklearn.pipeline import Pipeline

from datto.CleanText import CleanText


class ModelResults:
    def most_similar_texts(
        self, X, num_examples, text_column_name, num_topics=None, chosen_stopwords=set()
    ):
        """
        Uses NMF clustering to create n topics based on adjusted word frequencies

        Parameters
        --------
        X: DataFrame
        num_examples: int
        text_column_name: str
        num_topics: int
            Optional - if none algorithm will determine best number

        Returns
        --------
        topic_words_df: DataFrame
            Top 15 words/phrases per topic
        combined_df: DataFrame
            Original text with topic number assigned to each

        """
        X = X[~X[text_column_name].isna()]
        X = X[X[text_column_name] != ""]
        X = X[X[text_column_name] != " "]
        X = X[X[text_column_name] != "NA"]
        X = X[X[text_column_name] != "n/a"]
        X = X[X[text_column_name] != "N/A"]
        X = X[X[text_column_name] != "na"]

        all_stop_words = (
            set(ENGLISH_STOP_WORDS)
            | set(["-PRON-"])
            | set(string.punctuation)
            | set([" "])
            | chosen_stopwords
        )

        ct = CleanText()
        vectorizer = TfidfVectorizer(
            tokenizer=ct.lematize,
            ngram_range=(1, 3),
            stop_words=all_stop_words,
            min_df=5,
            max_df=0.4,
        )
        vectors = vectorizer.fit_transform(X[text_column_name]).todense()

        # Adding words/phrases used in text data frequencies back into the dataset (so we can see feature importances later)
        vocab = vectorizer.get_feature_names()
        vector_df = pd.DataFrame(vectors, columns=vocab, index=X.index)

        if X.shape[0] < 20:
            return "Too few examples to categorize."

        if not num_topics:

            # In case 1, add 1 to get at least 2
            # The rest are based on eyeballing numbers
            min_topics = ceil(X.shape[0] * 0.01) + 1
            max_topics = ceil(X.shape[0] * 0.2)
            step = ceil((max_topics - min_topics) / 5)

            topic_nums = list(np.arange(min_topics, max_topics, step))

            texts = X[text_column_name].apply(ct.lematize)

            # In gensim a dictionary is a mapping between words and their integer id
            dictionary = Dictionary(texts)

            # Filter out extremes to limit the number of features
            dictionary.filter_extremes(no_below=2, no_above=0.85, keep_n=5000)

            # Create the bag-of-words format (list of (token_id, token_count))
            corpus = [dictionary.doc2bow(text) for text in texts]

            coherence_scores = []

            for num in topic_nums:
                model = nmf.Nmf(
                    corpus=corpus,
                    num_topics=num,
                    id2word=dictionary,
                    chunksize=2000,
                    passes=5,
                    kappa=0.1,
                    minimum_probability=0.01,
                    w_max_iter=300,
                    w_stop_condition=0.0001,
                    h_max_iter=100,
                    h_stop_condition=0.001,
                    eval_every=10,
                    normalize=True,
                    random_state=42,
                )

                cm = CoherenceModel(
                    model=model, texts=texts, dictionary=dictionary, coherence="u_mass"
                )

                coherence_scores.append(round(cm.get_coherence(), 5))

            scores = list(zip(topic_nums, coherence_scores))
            chosen_num_topics = sorted(scores, key=itemgetter(1), reverse=True)[0][0]
        else:
            chosen_num_topics = num_topics

        model = NMF(n_components=chosen_num_topics, random_state=42)
        model.fit(vectors)
        component_loadings = model.transform(vectors)

        top_topics = pd.DataFrame(
            np.argmax(component_loadings, axis=1), columns=["top_topic_num"]
        )

        top_topic_loading = pd.DataFrame(
            np.max(component_loadings, axis=1), columns=["top_topic_loading"]
        )

        X.reset_index(inplace=True, drop=False)
        vector_df.reset_index(inplace=True, drop=True)

        # Fix for duplicate text_column_name
        vector_df.columns = [x + "_vector" for x in vector_df.columns]

        combined_df = pd.concat([X, vector_df, top_topics, top_topic_loading], axis=1)

        combined_df.sort_values(by="top_topic_loading", ascending=False, inplace=True)

        combined_df = pd.concat([X, vector_df, top_topics], axis=1)

        topic_words = {}
        sample_texts_lst = []
        for topic, comp in enumerate(model.components_):
            word_idx = np.argsort(comp)[::-1][:num_examples]
            topic_words[topic] = [vocab[i] for i in word_idx]
            sample_texts_lst.append(
                list(
                    combined_df[combined_df["top_topic_num"] == topic][
                        text_column_name
                    ].values[:num_examples]
                )
            )

        topic_words_df = pd.DataFrame(
            columns=[
                "topic_num",
                "num_in_category",
                "top_words_and_phrases",
                "sample_texts",
            ]
        )

        topic_words_df["topic_num"] = [k for k, _ in topic_words.items()]
        topic_words_df["num_in_category"] = (
            combined_df.groupby("top_topic_num").count().iloc[:, 0]
        )
        topic_words_df["top_words_and_phrases"] = [x for x in topic_words.values()]
        topic_words_df["sample_texts"] = sample_texts_lst

        topic_words_explode = pd.DataFrame(
            topic_words_df["sample_texts"].tolist(), index=topic_words_df.index,
        )

        topic_words_explode.columns = [
            "example{}".format(num) for num in range(len(topic_words_explode.columns))
        ]

        concated_topics = pd.concat(
            [
                topic_words_df[
                    ["topic_num", "num_in_category", "top_words_and_phrases"]
                ],
                topic_words_explode,
            ],
            axis=1,
        )

        print("Topics created with top words & example texts:")
        print(concated_topics)

        return (
            concated_topics,
            combined_df[["index", text_column_name, "top_topic_num"]],
            model
        )

    def coefficients_graph(self, X_train, X_test, model, model_type, filename):
        """
        Displays graph of feature importances.

        * Number of horizontal axis indicates magnitude of effect on
            target variable (e.g. affected by 0.25)
        * Red/blue indicates feature value (increasing or decreasing feature 
            has _ effect)
        * Blue & red mixed together indicate there isn't a clear 
            effect on the target variable
        * For classification - interpreting magnitude number / x axis - changes the 
            predicted probability of y on average by _ percentage points (axis value * 100)

        Parameters
        --------
        X_train: pd.DataFrame
        X_test: pd.DataFrame
        model: fit model object
        model_type: str
            'classification' or 'regression'
        filename: str
        """
        if model_type.lower() == "classification":
            f = lambda x: model.predict_proba(x)[:, 1]
        else:
            f = lambda x: model.predict(x)
        med = X_train.median().values.reshape((1, X_train.shape[1]))
        explainer = shap.KernelExplainer(f, med)
        # Runs too slow if X_test is huge, take a representative sample
        if X_test.shape[0] > 1000:
            X_test_sample = X_test.sample(1000)
        else:
            X_test_sample = X_test
        shap_values = explainer.shap_values(X_test_sample)
        shap.summary_plot(shap_values, X_test_sample)
        plt.tight_layout()
        plt.savefig(filename)

        return shap_values

    def most_common_words_by_group(
        self, X, text_col_name, group_col_name, num_examples, num_times_min, min_ngram,
    ):
        """
        Get the most commons phrases for defined groups.

        Parameters
        --------
        X: DataFrame
        text_col_name: str
        group_col_name: str
        num_examples: int
            Number of text examples to include per group
        num_times_min: int
            Minimum number of times word/phrase must appear in texts
        min_ngram: int

        Returns
        --------
        overall_counts_df: DataFrame
            Has groups, top words, and counts

        """
        # Fix for when column name is the same as an ngram column name
        X["group_column"] = X[group_col_name]

        # Remove all other unneeded columns
        X = X[[text_col_name, "group_column"]]

        all_stop_words = (
            set(ENGLISH_STOP_WORDS)
            | set(["-PRON-"])
            | set(string.punctuation)
            | set([" "])
        )

        cv = CountVectorizer(
            stop_words=all_stop_words,
            ngram_range=(min_ngram, 3),
            min_df=num_times_min,
            max_df=0.4,
        )
        vectors = cv.fit_transform(X[text_col_name]).todense()
        words = cv.get_feature_names()
        vectors_df = pd.DataFrame(vectors, columns=words)

        group_plus_vectors = pd.concat([vectors_df, X.reset_index(drop=False)], axis=1)

        count_words = pd.DataFrame(
            group_plus_vectors.groupby("group_column").count()["index"]
        )
        count_words = count_words.loc[:, ~count_words.columns.duplicated()]
        # Fix for when "count" is an ngram column
        count_words.columns = ["count_ngrams"]

        group_plus_vectors = group_plus_vectors.merge(
            count_words, on="group_column", how="left"
        )

        group_plus_vectors["count_ngrams"].fillna(0, inplace=True)

        sums_by_col = (
            group_plus_vectors[
                group_plus_vectors.columns[
                    ~group_plus_vectors.columns.isin([text_col_name, "index",])
                ]
            ]
            .groupby("group_column")
            .sum()
        )

        sums_by_col.sort_values(by="count_ngrams", ascending=False, inplace=True)

        sums_by_col.drop("count_ngrams", axis=1, inplace=True)

        array_sums = np.array(sums_by_col)
        sums_values_descending = -np.sort(-array_sums, axis=1)
        sums_indices_descending = (-array_sums).argsort()

        highest_sum = pd.DataFrame(sums_values_descending[:, 0])
        highest_sum.columns = ["highest_sum"]

        sums_by_col["highest_sum"] = highest_sum["highest_sum"].values

        overall_counts_df = pd.DataFrame(columns=["group_name", "top_words_and_counts"])
        i = 0
        for row in sums_by_col.index:
            dict_scores = {}
            temp_df = pd.DataFrame(columns=["group_name", "top_words_and_counts"])
            temp_df["group_name"] = [row]
            top_columns = sums_by_col.columns[
                sums_indices_descending[i][:num_examples]
            ].values
            top_counts = sums_values_descending[i][:num_examples]
            [dict_scores.update({x: y}) for x, y in zip(top_columns, top_counts)]
            temp_df["top_words_and_counts"] = [dict_scores]
            overall_counts_df = overall_counts_df.append([temp_df])
            print(f"Group Name: {row}\n")
            for k, v in dict_scores.items():
                print(k, v)
            print("\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n")
            i += 1

        return overall_counts_df

    def score_final_model(
        self, model_type, X_test, y_test, trained_model, multiclass=False
    ):
        """
        Score your model on the test dataset. Only run this once to get an idea of how your model will perform in realtime.
        Run it after you have chosen your model & parameters to avoid problems with overfitting.

        Parameters
        --------
        model_type: str
        X_test: DataFrame
        y_test: DataFrame
        trained_model: sklearn model
        multiclass: bool

        Returns
        --------
        model: model
            Fit model
        y_predicted: array
        """
        # Predict actual scores
        y_predicted = trained_model.predict(X_test)

        if multiclass:
            pscore = precision_score(y_test, y_predicted, average="weighted")
            rscore = recall_score(y_test, y_predicted, average="weighted")
            f1score = f1_score(y_test, y_predicted, average="weighted")

            print(f"Final Model Precision Weighted: {pscore}")
            print(f"Final Model Recall Weighted: {rscore}")
            print(f"Final Model F1 Weighted: {f1score}")
        elif model_type.lower() == "classification":
            pscore = precision_score(y_test, y_predicted)
            rscore = recall_score(y_test, y_predicted)
            accuracy = accuracy_score(y_test, y_predicted)
            roc_auc = roc_auc_score(y_test, y_predicted)

            print(f"Final Model Precision: {pscore}")
            print(f"Final Model Recall: {rscore}")
            print(f"Final Model Accuracy: {accuracy}")
            print(f"Final Model ROC AUC: {roc_auc}")

            crosstab = pd.crosstab(
                y_test, y_predicted, rownames=["Actual"], colnames=["Predicted"],
            )
            print(crosstab)
            sum_crosstab = crosstab.to_numpy().sum()
            print(
                pd.crosstab(
                    y_test, y_predicted, rownames=["Actual"], colnames=["Predicted"],
                ).apply(lambda r: round(r / sum_crosstab, 3))
            )
        else:
            mse = mean_squared_error(y_test, y_predicted)
            mae = median_absolute_error(y_test, y_predicted)
            r2 = r2_score(y_test, y_predicted)

            print(f"Mean Negative Root Mean Squared Errror: {(mse ** 5) * -1}")
            print(f"Mean Negative Median Absolute Error: {mae * -1}")
            print(f"Mean R2: {r2}")

        return trained_model, y_predicted

    def coefficients_summary(
        self,
        X,
        y,
        num_repetitions,
        num_coefficients,
        model_type,
        multiclass=False,
        params={},
    ):
        """
        Prints average coefficient values using a regression model.

        Parameters
        --------
        X: DataFrame
        y: DataFrame
        num_repetitions: int
            Number of times to create models
        num_coefficients: int
            Number of top coefficients to display
        model_type: str
            'classification' or 'regression'
        multiclass: bool
        params: dict
            Optional - add to change regression params, otherwise use default
            
        Returns
        --------
        simplified_df: DataFrame
            Has mean, median, and standard deviation for coefficients after several runs
        """
        coefficients_df = pd.DataFrame(columns=["features", "coefficients"])
        X["intercept"] = 1

        for _ in range(num_repetitions):
            if multiclass:
                model = DecisionTreeClassifier()
            elif model_type.lower() == "classification":
                model = LogisticRegression(fit_intercept=False, **params)
            else:
                model = ElasticNet(fit_intercept=False, **params)

            X_train, _, y_train, _ = train_test_split(X, y)

            model.fit(X_train, y_train)

            if multiclass:
                coefs = model.feature_importances_
            elif model_type.lower() == "classification":
                coefs = model.coef_[0]
            else:
                coefs = model.coef_

            temp_df = pd.DataFrame(
                [x for x in zip(X_train.columns, coefs)],
                columns=["features", "coefficients"],
            )

            coefficients_df = coefficients_df.append(temp_df)

        column_of_interest = "coefficients"

        summary_coefficients_df = pd.DataFrame(
            coefficients_df.groupby("features").agg(
                {column_of_interest: ["mean", "std", "median"]}
            )
        )

        summary_coefficients_df.columns = ["mean", "std", "median"]

        summary_coefficients_df.reset_index(inplace=True)

        value_counts_df = pd.DataFrame(columns=["features"])
        for col in X_train.columns:
            temp_df = pd.DataFrame([[col,]], columns=["features"],)
            value_counts_df = value_counts_df.append(temp_df)
        value_counts_df.reset_index(inplace=True, drop=True)

        combined_df = summary_coefficients_df.merge(
            value_counts_df, on="features", how="left"
        )

        combined_df["abs_val_mean"] = abs(combined_df["mean"])
        combined_df["abs_val_to_se"] = abs(combined_df["mean"]) / combined_df["std"]

        combined_df.sort_values("abs_val_to_se", inplace=True, ascending=False)

        simplified_df = (
            combined_df[["features", "mean", "std", "median", "abs_val_to_se"]]
            .head(num_coefficients)
            .round(7)
        )

        # 7 is the most before it flips back to scientific notation
        print("Coefficients summary (descending by mean abs se value):")
        print(simplified_df)

        return simplified_df

    def coefficients_individual_predictions(
        self,
        trained_model,
        X_train,
        X_test,
        id_col,
        num_samples,
        model_type,
        class_names=["False", "True"],
    ):
        def model_preds_adjusted(data):
            if model_type.lower() == "classification":
                predictions = np.array(trained_model.predict_proba(data))
            else:
                predictions = np.array(trained_model.predict(data))
            return predictions

        if model_type.lower() == "classification":
            explainer = lime.lime_tabular.LimeTabularExplainer(
                np.array(X_train),
                feature_names=X_train.columns,
                class_names=class_names,
                mode="classification",
            )
        else:
            explainer = lime.lime_tabular.LimeTabularExplainer(
                np.array(X_train),
                feature_names=X_train.columns,
                class_names=class_names,
                mode="regression",
            )

        for i in range(num_samples):
            user_idx = randrange(X_test.shape[0])
            exp = explainer.explain_instance(
                np.array(X_test.iloc[user_idx]),
                model_preds_adjusted,
                num_features=50,
                top_labels=10,
            )

            user_id = X_test.iloc[user_idx][id_col]
            print(f"\nUser: {user_id}")

            if model_type.lower() == "classification":
                prediction = class_names[
                    trained_model.predict_proba(
                        pd.DataFrame(X_test.iloc[user_idx]).T
                    ).argmax()
                ]
            else:
                prediction = round(
                    trained_model.predict(pd.DataFrame(X_test.iloc[user_idx]).T)[0], 7
                )

            print('We predicted "{}" for this user because...\n'.format(prediction))
            features_list = []
            for feature in exp.as_list():
                features_list.append({user_id: exp.as_list()})
                try:
                    feature_name = re.findall("<.*<|>.*>", feature[0])[0]
                except Exception:
                    feature_name = re.findall(".*<|.*>", feature[0])[0]
                cleaned_feature_name = (
                    feature_name.replace("<=", "")
                    .replace(">=", "")
                    .replace("<", "")
                    .replace(">", "")
                    .strip()
                )
                comparison_val = float(
                    feature[0].split(" <")[-1].split(" >")[-1].split("=")[-1].strip()
                )
                actual_feature_val = X_test.iloc[user_idx][cleaned_feature_name]
                last_operator = feature[0].split(" ")[-2]

                if "<" in last_operator and actual_feature_val <= comparison_val:
                    print(" * This is true: " + feature[0])
                else:
                    print(" * This is false: " + feature[0])
            exp.as_pyplot_figure()
            plt.tight_layout()
            plt.savefig(f"lime_graph_user_{user_id}.png")

            print("\n\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n")

        return features_list




import os
import re
import string
from collections import Counter

import numpy as np
import pandas as pd
import spacy
from spacy.cli import download

download("en")
nlp = spacy.load("en")


class CleanText:
    def remove_names(self, text):
        """
        Parameters
        --------
        text: str

        Returns
        --------
        cleaned_text: str

        """
        all_names = pd.read_pickle(
            os.path.join(os.path.dirname(__file__), "data/all_names")
        )
        cleaned_text = text
        for _, row in all_names.iterrows():
            # Matches name as long as it is not followed by lowercase characters
            # Removing names that are a part of another word
            cleaned_text = re.sub(row["name"] + "(?![a-z])", " ", cleaned_text)
        return cleaned_text

    def remove_links(self, text):
        """
        Parameters
        --------
        text: str

        Returns
        --------
        cleaned_text: str

        """
        cleaned_text = text
        links_found = re.findall(
            r"(https?:\/\/(?:www\.|(?!www))[a-zA-Z0-9][a-zA-Z0-9-]+[a-zA-Z0-9]\.[^\s]{2,}|www\.[a-zA-Z0-9][a-zA-Z0-9-]+[a-zA-Z0-9]\.[^\s]{2,}|https?:\/\/(?:www\.|(?!www))[a-zA-Z0-9]\.[^\s]{2,}|www\.[a-zA-Z0-9]\.[^\s]{2,})",
            cleaned_text,
        )
        for link in links_found:
            cleaned_text = cleaned_text.replace(link, "")
        return cleaned_text

    def lematize(self, text):
        """
        Parameters
        --------
        text: str

        Returns
        --------
        list of spacy tokens

        """
        spacy_text = nlp(text)
        return [token.lemma_ for token in spacy_text if not token.is_space]

    def remove_email_greetings_signatures(self, text):
        """
        In order to obtain the main text of an email only, this method removes greetings, signoffs,
        and signatures by identifying sentences with less than 5% verbs to drop. Does not replace links.

        Inspiration from: https://github.com/mynameisvinn/EmailParser

        Parameters
        --------
        text: str

        Returns
        --------
        text: str

        """
        sentences = text.strip().split("\n")
        non_sentences = []

        for sentence in sentences:
            spacy_text = nlp(sentence.strip())
            verb_count = np.sum(
                [
                    (
                        token.pos_ == "VERB"
                        or token.pos_ == "AUX"
                        or token.pos_ == "ROOT"
                        or token.pos_ == "pcomp"
                    )
                    for token in spacy_text
                ]
            )
            try:
                prob = float(verb_count) / len(spacy_text)
            except Exception:
                prob = 1.0

            # If 5% or less of a sentence is verbs, it's probably not a real sentence
            if prob <= 0.05:
                non_sentences.append(sentence)

        for non_sentence in non_sentences:
            # Don't replace links
            if "http" not in non_sentence and non_sentence not in string.punctuation:
                text = text.replace(non_sentence, "")

        return text

    def clean_column_names(self, df):
        """
        Rename all columns to use underscores to reference columns without bracket formatting

        Parameters
        --------
        df: DataFrame

        Returns
        --------
        df: DataFrame

        """
        df.rename(
            columns=lambda x: str(x).strip().replace(" ", "_").lower(), inplace=True
        )
        return df

    def remove_duplicate_columns(self, df):
        """
        Remove columns with the same name

        Parameters
        --------
        df: DataFrame

        Returns
        --------
        df: DataFrame

        """
        df = df.loc[:, ~df.columns.duplicated()]
        return df

    def fix_col_data_type(self, df, col, desired_dt):
        """
        Change column datatype using the best method for each type.

        Parameters
        --------
        df: DataFrame
        col: str
            Column to change the dtype for
        desired_dt: str
            {'float', 'int', 'datetime', 'str'}

        Returns
        --------
        df: DataFrame

        """
        if desired_dt in ("float", "int"):
            df[col] = pd.to_numeric(df[col], errors="coerce")
        elif desired_dt == "datetime":
            df[col] = pd.to_datetime(df[col], errors="coerce")
        elif desired_dt == "str":
            df[col] = df[col].astype(str)

        return df

    def compress_df(self, df):
        """
        Compresses each dataframe column as much as possible depending on type and values.

        Parameters
        --------
        df: DataFrame

        Returns
        --------
        df: DataFrame

        """
        for col in df.columns:
            if df[col].dtype == "O":
                unique_vals = df[col].nunique()
                count_vals = df[col].shape[0]
                if unique_vals < (count_vals * 0.5):
                    df[col] = df[col].astype("category")
            elif df[col].dtype == "int64":
                df[col] = pd.to_numeric(df[col], downcast="unsigned")
            elif df[col].dtype == "float64":
                df[col] = pd.to_numeric(df[col], downcast="float")

        return df

    def make_uuid(self, id):
        """
        Make a UUID from a text string

        Parameters
        --------
        id: str

        Returns
        --------
        uuid: str

        """
        return (
            id[:8] + "-" + id[8:12] + "-" + id[12:16] + "-" + id[16:20] + "-" + id[20:]
        )

    def df_most_common_only(self, df, col, num):
        """
        From a given column in a dataframe, find the most common values and return a dataframe that has those most common values only.

        Parameters
        --------
        df: DataFrame
        col: str
        num: int

        Returns
        --------
        cleaned_df: DataFrame

        """
        most_common_lst = [x[0] for x in Counter(df[col]).most_common(num)]
        df = df[df[col].isin(most_common_lst)]
        return df

    def batch_pandas_operation(self, df, num_splits, identifier_col, func):
        """
        Use a function on a Pandas DataFrame in chunks for faster processing.

        Parameters
        --------
        df: DataFrame
        num_splits: int
        identifier_col: str
        func: Function

        Returns
        --------
        new_df: DataFrame

        """
        new_df = pd.DataFrame()
        ids = df[identifier_col].unique()
        division = len(ids) / float(num_splits)
        part_ids = [
            ids[int(round(division * i)) : int(round(division * (i + 1)))]
            for i in range(num_splits)
        ]
        for lst in part_ids:
            temp_df = df[df[identifier_col].isin(lst)]
            new_df = new_df.append(func(temp_df))

        return new_df

    def batch_merge_operation(self, df_1, df_2, num_splits, identifier_col, merge_col):
        """
        Merge two Pandas DataFrame in chunks for faster processing.

        Parameters
        --------
        df_1: DataFrame
        df_2: DataFrame
        num_splits: int
        identifier_col: str
        merge_col: str

        Returns
        --------
        new_df: DataFrame

        """
        new_df = pd.DataFrame()
        ids = df_1[identifier_col].unique()
        division = len(ids) / float(num_splits)
        part_ids = [
            ids[int(round(division * i)) : int(round(division * (i + 1)))]
            for i in range(num_splits)
        ]
        for lst in part_ids:
            temp_df = df_1[df_1[identifier_col].isin(lst)]
            temp_df = pd.merge(temp_df, df_2, how="left", on=merge_col)
            new_df = new_df.append(temp_df)
        return new_df



import pandas as pd


class Eda:
    def separate_cols_by_type(self, df):
        """
        Split the DataFrame into two groups by type

        Parameters
        --------
        df: DataFrame

        Returns
        --------
        numerical_vals: DataFrame
        categorical_vals: DataFrame
        """
        numerical_vals = df.select_dtypes(exclude=["object", "bool"])
        categorical_vals = df.select_dtypes(include=["object", "bool"])

        return numerical_vals, categorical_vals

    def check_for_mistyped_booleans(self, numerical_vals):
        """
        Check for columns coded as ints/floats that should actually be booleans

        Parameters
        --------
        numerical_vals: list

        Returns
        --------
        boolean_cols: list

        """
        boolean_cols = []
        for col in numerical_vals:
            if numerical_vals[col].nunique() <= 2:
                print(col)
                print(numerical_vals[col].unique())
                boolean_cols.append(col)

        return boolean_cols

    def find_cols_to_exclude(self, df):
        """
        Returns columns that may not be helpful for model building.

        Exclusion criteria:
        - Possible PII (address, name, username, date, etc. in col name)
        - Large proportion of nulls
        - Only 1 value in entire col
        - Dates
        - Low variance in col values
        - Large number of categorical values

        Parameters
        --------
        df: DataFrame

        Returns
        --------
        lst: list

        """
        lst = []
        for col in df.columns:
            if (
                "address" in str(col)
                or "first_name" in str(col)
                or "last_name" in str(col)
                or "username" in str(col)
                or "_id" in str(col)
                or "date" in str(col)
                or "time" in str(col)
            ):
                lst.append({col: "Considering excluding because potential PII column."})
            elif df[col].isnull().sum() / float(df.shape[0]) >= 0.5:
                lst.append(
                    {
                        col: "Considering excluding because {}% of column is null.".format(
                            (df[col].isnull().sum() / float(df.shape[0]) * 100.0)
                        )
                    }
                )
            elif len(df[col].unique()) <= 1:
                lst.append(
                    {
                        col: "Considering excluding because column includes only one value."
                    }
                )
            elif df[col].dtype == "datetime64[ns]":
                lst.append(
                    {col: "Considering excluding because column is a timestamp."}
                )
            elif df[col].dtype not in ["object", "bool"]:
                if df[col].var() < 0.00001:
                    lst.append(
                        {
                            col: "Considering excluding because column variance is low ({})".format(
                                df[col].var()
                            )
                        }
                    )
            elif df[col].dtype in ["object", "bool"]:
                if len(df[col].unique()) > 200:
                    lst.append(
                        {
                            col: "Considering excluding because object column has large number of unique values ({})".format(
                                len(df[col].unique())
                            )
                        }
                    )

        [print(x) for x in lst]

        return lst

    def sample_unique_vals(self, df):
        """
        Examine a few unique vals in each column

        Parameters
        --------
        df: DataFrame
        """
        for col in df:
            print(col)
            try:
                print(df[col].unique()[:20])
                print(df[col].nunique())
            except Exception:
                pass
            print("\n------------------------------------\n")

    def find_correlated_features(self, df):
        """
        Find & sort correlated features

        Parameters
        --------
        df: DataFrame

        Returns
        --------
        s: Series

        """
        if df.empty:
            return pd.Series()

        c = df.corr().abs()
        s = c.unstack()
        s = s[s <= 0.99999]
        s = s.sort_values(ascending=False)
        print(s)
        return s



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['datto', 'datto.data']

package_data = \
{'': ['*']}

install_requires = \
['black>=19.10b0,<20.0',
 'catboost>=0.24.1,<0.25.0',
 'gensim>=3.8.3,<4.0.0',
 'hypothesis>=5.35.2,<6.0.0',
 'ipython>=7.8,<8.0',
 'lightgbm>=3.0.0,<4.0.0',
 'lime>=0.2.0,<0.3.0',
 'notebook>=6.0.3,<7.0.0',
 'pandas>=1.0.4,<2.0.0',
 'psycopg2-binary>=2.8.5,<3.0.0',
 'pylint>=2.5.0,<3.0.0',
 'pytest>=5.2,<6.0',
 'python-json-logger>=0.1.11,<0.2.0',
 's3fs>=0.4.2,<0.5.0',
 'shap>=0.36.0,<0.37.0',
 'sklearn>=0.0,<0.1',
 'spacy>=2.2.4,<3.0.0',
 'sphinx-rtd-theme>=0.5.0,<0.6.0',
 'sphinx>=3.2.1,<4.0.0',
 'xgboost>=1.1.1,<2.0.0']

setup_kwargs = {
    'name': 'datto',
    'version': '0.4.7',
    'description': 'Data Tools (Dat To)',
    'long_description': None,
    'author': 'kristiewirth',
    'author_email': 'kristie.ann.wirth@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)
